from argparse import ArgumentParser, _SubParsersAction


def create_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    ...


def edit_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    ...


def get_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    ...


def delete_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    ...



import asyncio
import io
import os
import shutil
import traceback
import uuid
from pathlib import Path

from fastapi import APIRouter, Request, Response
from loguru import logger

from pipeline.cloud.http import StreamingResponseWithStatusCode
from pipeline.cloud.schemas import pipelines as pipeline_schemas
from pipeline.cloud.schemas import runs as run_schemas
from pipeline.container.manager import Manager
from pipeline.exceptions import RunInputException, RunnableError
from pipeline.objects.graph import File

router = APIRouter(prefix="/runs", tags=["runs"])


@router.post(
    "",
    status_code=200,
    responses={
        500: {
            "description": "Pipeline failed",
            "model": run_schemas.ContainerRunResult,
        },
        400: {
            "description": "Invalid input data",
            "model": run_schemas.ContainerRunResult,
        },
    },
)
async def run(
    run_create: run_schemas.ContainerRunCreate,
    request: Request,
    response: Response,
) -> run_schemas.ContainerRunResult:
    run_id = run_create.run_id
    with logger.contextualize(run_id=run_id):
        manager: Manager = request.app.state.manager
        if result := _handle_pipeline_state_not_ready(manager):
            return result

        execution_queue: asyncio.Queue = request.app.state.execution_queue

        response_queue: asyncio.Queue = asyncio.Queue()
        execution_queue.put_nowait((run_create, response_queue))
        run_output = await response_queue.get()
        logger.info("Returning run result")
        response_schema, response.status_code = _generate_run_result(run_output)
        return response_schema


@router.post(
    "/stream",
    status_code=200,
    responses={
        500: {
            "description": "Pipeline failed",
            "model": run_schemas.ContainerRunResult,
        },
        400: {
            "description": "Invalid input data",
            "model": run_schemas.ContainerRunResult,
        },
    },
)
async def stream_run(
    run_create: run_schemas.ContainerRunCreate,
    request: Request,
    response: Response,
):
    run_id = run_create.run_id
    with logger.contextualize(run_id=run_id):
        manager: Manager = request.app.state.manager
        if result := _handle_pipeline_state_not_ready(manager):
            return result

        execution_queue: asyncio.Queue = request.app.state.execution_queue

        response_queue: asyncio.Queue = asyncio.Queue()
        execution_queue.put_nowait((run_create, response_queue))
        run_output = await response_queue.get()

        response_schema, response.status_code = _generate_run_result(run_output)

        outputs = response_schema.outputs or []
        if not outputs:
            return response_schema

        if not any([output.type == run_schemas.RunIOType.stream for output in outputs]):
            raise TypeError("No streaming outputs found")

        return StreamingResponseWithStatusCode(
            _stream_run_outputs(response_schema, request),
            media_type="application/json",
            # hint to disable buffering
            headers={"X-Accel-Buffering": "no"},
        )


def _fetch_next_outputs(outputs: list[run_schemas.RunOutput]):
    next_outputs = []
    have_new_streamed_outputs = False
    for output in outputs:
        if output.type == run_schemas.RunIOType.stream:
            if output.value is None:
                raise Exception("Stream value was None")

            try:
                next_value = output.value.__next__()
                next_outputs.append(
                    run_schemas.RunOutput(
                        type=run_schemas.RunIOType.from_object(next_value),
                        value=next_value,
                        file=None,
                    )
                )
                have_new_streamed_outputs = True
            except StopIteration:
                # if no data left for this stream, return None value
                next_outputs.append(
                    run_schemas.RunOutput(
                        type=run_schemas.RunIOType.none,
                        value=None,
                        file=None,
                    )
                )
            except Exception as exc:
                logger.exception("Pipeline error caught during run streaming")
                raise RunnableError(exception=exc, traceback=traceback.format_exc())
        else:
            next_outputs.append(output)

    if not have_new_streamed_outputs:
        return
    return next_outputs


async def _stream_run_outputs(
    response_schema: run_schemas.ContainerRunResult, request: Request
):
    """Generator returning output data for list of outputs.

    We iterate over all outputs until we no longer have any streamed data to
    output
    """
    outputs = response_schema.outputs or []
    while True:
        status_code = 200
        try:
            next_outputs = _fetch_next_outputs(outputs)
            if not next_outputs:
                return

            new_response_schema = run_schemas.ContainerRunResult(
                inputs=response_schema.inputs,
                outputs=next_outputs,
                error=response_schema.error,
            )

        except RunnableError as e:
            # if we get a pipeline error, return a run error then finish
            new_response_schema = run_schemas.ContainerRunResult(
                outputs=None,
                inputs=response_schema.inputs,
                error=run_schemas.ContainerRunError(
                    type=run_schemas.ContainerRunErrorType.pipeline_error,
                    message=repr(e.exception),
                    traceback=e.traceback,
                ),
            )
        except Exception as e:
            logger.exception("Unexpected error during run streaming")
            status_code = 500
            new_response_schema = run_schemas.ContainerRunResult(
                outputs=None,
                inputs=response_schema.inputs,
                error=run_schemas.ContainerRunError(
                    type=run_schemas.ContainerRunErrorType.unknown,
                    message=repr(e),
                    traceback=None,
                ),
            )

        # serialise response to str and add newline separator
        yield f"{new_response_schema.json()}\n", status_code

        # if there was an error or request is disconnected terminate all iterators
        if new_response_schema.error or await request.is_disconnected():
            for output in outputs:
                if (
                    output.type == run_schemas.RunIOType.stream
                    and output.value is not None
                    and hasattr(output.value.iterable, "end")
                ):
                    output.value.iterable.end()
            return


def _handle_pipeline_state_not_ready(
    manager: Manager,
) -> run_schemas.ContainerRunResult | None:
    if manager.pipeline_state == pipeline_schemas.PipelineState.loading:
        logger.info("Pipeline loading")
        return run_schemas.ContainerRunResult(
            outputs=None,
            inputs=None,
            error=run_schemas.ContainerRunError(
                type=run_schemas.ContainerRunErrorType.pipeline_loading,
                message="Pipeline is still loading",
                traceback=None,
            ),
        )

    if manager.pipeline_state == pipeline_schemas.PipelineState.load_failed:
        logger.info("Pipeline failed to load")
        return run_schemas.ContainerRunResult(
            outputs=None,
            inputs=None,
            error=run_schemas.ContainerRunError(
                type=run_schemas.ContainerRunErrorType.startup_error,
                message="Pipeline failed to load",
                traceback=manager.pipeline_state_message,
            ),
        )

    if manager.pipeline_state == pipeline_schemas.PipelineState.startup_failed:
        logger.info("Pipeline failed to startup")
        return run_schemas.ContainerRunResult(
            outputs=None,
            inputs=None,
            error=run_schemas.ContainerRunError(
                type=run_schemas.ContainerRunErrorType.startup_error,
                message="Pipeline failed to startup",
                traceback=manager.pipeline_state_message,
            ),
        )


def _generate_run_result(run_output) -> tuple[run_schemas.ContainerRunResult, int]:
    if isinstance(run_output, RunInputException):
        return (
            run_schemas.ContainerRunResult(
                outputs=None,
                inputs=None,
                error=run_schemas.ContainerRunError(
                    type=run_schemas.ContainerRunErrorType.input_error,
                    message=run_output.message,
                    traceback=None,
                ),
            ),
            400,
        )
    elif isinstance(run_output, RunnableError):
        return (
            run_schemas.ContainerRunResult(
                outputs=None,
                inputs=None,
                error=run_schemas.ContainerRunError(
                    type=run_schemas.ContainerRunErrorType.pipeline_error,
                    message=repr(run_output.exception),
                    traceback=run_output.traceback,
                ),
            ),
            200,
        )
    elif isinstance(run_output, Exception):
        return (
            run_schemas.ContainerRunResult(
                outputs=None,
                inputs=None,
                error=run_schemas.ContainerRunError(
                    type=run_schemas.ContainerRunErrorType.unknown,
                    message=str(run_output),
                    traceback=None,
                ),
            ),
            500,
        )
    else:
        outputs = _parse_run_outputs(run_output)
        return (
            run_schemas.ContainerRunResult(
                outputs=outputs,
                error=None,
                inputs=None,
            ),
            200,
        )


def _parse_run_outputs(run_outputs):
    outputs = []
    for output in run_outputs:
        output_type = run_schemas.RunIOType.from_object(output)
        # if single file
        if output_type == run_schemas.RunIOType.file:
            file_schema = _save_run_file(output)
            outputs.append(
                run_schemas.RunOutput(type=output_type, value=None, file=file_schema)
            )
        # else if list of files
        elif (
            output_type == run_schemas.RunIOType.pkl
            and isinstance(output, list)
            and all([isinstance(item, (File, io.BufferedIOBase)) for item in output])
        ):
            file_list = []
            for file in output:
                file_schema = _save_run_file(file)
                file_list.append(
                    run_schemas.RunOutput(
                        type=run_schemas.RunIOType.file,
                        value=None,
                        file=file_schema,
                    )
                )
            outputs.append(
                run_schemas.RunOutput(
                    type=run_schemas.RunIOType.array,
                    value=file_list,
                    file=None,
                )
            )
        else:
            outputs.append(
                run_schemas.RunOutput(type=output_type, value=output, file=None)
            )
    return outputs


def _save_run_file(file: File | io.BufferedIOBase) -> run_schemas.RunOutputFile:
    # ensure we save the file somewhere unique on the file system so it doesn't
    # get overwritten by another run
    uuid_path = str(uuid.uuid4())
    output_path = Path(f"/tmp/run_files/{uuid_path}")
    output_path.mkdir(parents=True, exist_ok=True)

    if isinstance(file, File):
        # should always exist
        assert file.path
        file_name = file.path.name
        output_file = output_path / file_name
        file_size = os.stat(file.path).st_size
        # copy file to new output location
        shutil.copyfile(file.path, output_file)
    else:
        file_name = getattr(
            file,
            "name",
            str(uuid.uuid4()),
        )
        try:
            file_size = file.seek(0, os.SEEK_END)
        except Exception:
            file_size = -1
            logger.warning(f"Could not get size of type {type(file)}")
        # write file to new output location
        output_file = output_path / file_name
        output_file.write_bytes(file.read())

    file_schema = run_schemas.RunOutputFile(
        name=file_name, path=str(output_file), size=file_size, url=None
    )
    return file_schema



import contextlib
import inspect
import json
import logging
import os
import sys
import traceback

from loguru import logger


class StreamToLogger:
    def __init__(self, level="INFO"):
        self._level = level
        self.encoding = "utf-8"

    def write(self, buffer):
        for line in buffer.rstrip().splitlines():
            logger.opt(depth=1).log(self._level, line.rstrip())

    def flush(self):
        pass


@contextlib.contextmanager
def redirect_stdout():
    """Contextmanager for redirecting stdout to core logger.

    For more info see:
    https://loguru.readthedocs.io/en/stable/resources/recipes.html#capturing-standard-stdout-stderr-and-warnings
    """
    stream = StreamToLogger()
    with contextlib.redirect_stdout(stream):
        yield


class InterceptHandler(logging.Handler):
    """
    Ensure standard logging calls get forwarded to core Loguru logger
    https://loguru.readthedocs.io/en/stable/overview.html#entirely-compatible-with-standard-logging
    """

    def emit(self, record: logging.LogRecord) -> None:
        # Get corresponding Loguru level if it exists.
        level: str | int
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message.
        frame, depth = inspect.currentframe(), 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(
            level, record.getMessage()
        )


def default_log_handler(message, file=sys.stderr):
    # No `end` as `message` already has a newline
    print(message, file=file, end="")


def json_log_handler(message, file=sys.stderr):
    record = message.record
    payload = {
        "log.level": record["level"].name,
        "log.time.timestamp": record["time"].timestamp(),
        "log.time.iso8601": record["time"].isoformat(),
        "log.source.file": f"{record['file'].name}:{record['line']}",
        "log.source.module": record["name"],
        "log.message": record["message"],
        **record["extra"],
    }
    if (ex := record["exception"]) is not None:
        payload["exception.type"] = repr(ex.type)
        payload["exception.value"] = repr(ex.value)
        payload["exception.traceback"] = "".join(
            traceback.format_exception(ex.type, ex.value, ex.traceback, limit=200)
        )
    print(json.dumps(payload, default=str, ensure_ascii=False), file=file)


def setup_logging():
    """Setup logging:
    - Suppress uvicorn logs
    - Use loguru as default logging library
    - If USE_JSON_LOGGING is set, write JSON logs (these can then be queried and
        returned by main API), otherwise use default log format
    - Use InterceptHandler to forward all logs to loguru
    """
    # suppress uvicorn logs
    logging.getLogger("uvicorn").setLevel(logging.CRITICAL)
    logging.getLogger("uvicorn.error").setLevel(logging.CRITICAL)
    logging.getLogger("uvicorn.access").handlers.clear()
    logger.remove()

    use_json_logging = os.environ.get("USE_JSON_LOGGING", False)
    if use_json_logging:
        handler = dict(
            sink=json_log_handler,
            colorize=False,
        )
    else:
        handler = dict(
            sink=default_log_handler,
            colorize=True,
        )
    logger.configure(handlers=[handler])
    logging.basicConfig(handlers=[InterceptHandler()], level=0, force=True)



import asyncio
import os
import traceback
import uuid

import pkg_resources
from fastapi import FastAPI, Request
from fastapi.concurrency import run_in_threadpool
from fastapi.middleware.cors import CORSMiddleware
from fastapi.openapi.utils import get_openapi
from fastapi.responses import JSONResponse
from fastapi.staticfiles import StaticFiles
from loguru import logger

from pipeline.cloud.schemas import runs as run_schemas
from pipeline.container.logging import redirect_stdout, setup_logging
from pipeline.container.manager import Manager
from pipeline.container.routes import router
from pipeline.container.status import router as status_router


def create_app() -> FastAPI:
    app = FastAPI(
        title="pipeline-container",
    )

    setup_logging()

    setup_oapi(app)
    setup_middlewares(app)

    app.state.execution_queue = asyncio.Queue()
    app.state.manager = Manager(
        pipeline_path=os.environ.get(
            "PIPELINE_PATH",
            "",
        )
    )
    asyncio.create_task(execution_handler(app.state.execution_queue, app.state.manager))

    app.include_router(router)
    app.include_router(status_router)
    static_dir = pkg_resources.resource_filename(
        "pipeline", "container/frontend/static"
    )

    app.mount(
        "/static",
        StaticFiles(directory=static_dir),
        name="static",
    )

    return app


def setup_middlewares(app: FastAPI) -> None:
    @app.middleware("http")
    async def _(request: Request, call_next):
        request.state.request_id = request.headers.get("X-Request-Id") or str(
            uuid.uuid4()
        )

        try:
            response = await call_next(request)
            response.headers["X-Request-Id"] = request.state.request_id
        except Exception as e:
            logger.exception(e)
            response = JSONResponse(
                status_code=500,
                content={
                    "error": repr(e),
                    "traceback": str(traceback.format_exc()),
                },
            )
            response.headers["X-Request-Id"] = request.state.request_id
        return response

    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )


def setup_oapi(app: FastAPI) -> None:
    def custom_openapi():
        if app.openapi_schema:
            return app.openapi_schema
        openapi_schema = get_openapi(
            title="pipeline-container",
            version="1.1.0",
            routes=app.routes,
            servers=[{"url": "http://localhost:14300"}],
        )
        app.openapi_schema = openapi_schema
        return app.openapi_schema

    app.openapi = custom_openapi


async def execution_handler(execution_queue: asyncio.Queue, manager: Manager) -> None:
    with redirect_stdout():
        await run_in_threadpool(manager.startup)

        while True:
            try:
                args, response_queue = await execution_queue.get()
                args: run_schemas.ContainerRunCreate
                input_data = args.inputs
                run_id = args.run_id
                with logger.contextualize(run_id=run_id):
                    try:
                        output = await run_in_threadpool(
                            manager.run, run_id=run_id, input_data=input_data
                        )
                    except Exception as e:
                        logger.exception("Exception raised during pipeline execution")
                        response_queue.put_nowait(e)
                        continue
                    response_queue.put_nowait(output)
            except Exception:
                logger.exception("Got an error in the execution loop handler")



import pkg_resources
from fastapi import APIRouter
from fastapi.responses import FileResponse

router = APIRouter(prefix="/play", tags=["play"])


@router.get("", response_class=FileResponse)
async def serve_root():
    file_path = pkg_resources.resource_filename(
        "pipeline", "container/frontend/static/index.html"
    )
    return FileResponse(file_path)



import subprocess
from argparse import Namespace
from pathlib import Path

import docker
import docker.errors
import yaml
from docker.types import DeviceRequest, LogConfig

from pipeline.util.logging import _print

from .schemas import PipelineConfig


def up_container(namespace: Namespace):
    _print("Starting container...", "INFO")
    config_file = Path(getattr(namespace, "file", "./pipeline.yaml"))

    if not config_file.exists():
        raise FileNotFoundError(f"Config file {config_file} not found")

    config = config_file.read_text()
    pipeline_config_yaml = yaml.load(config, Loader=yaml.FullLoader)

    pipeline_config = PipelineConfig.parse_obj(pipeline_config_yaml)

    pipeline_name = pipeline_config.pipeline_name
    docker_client = docker.from_env()
    lc = LogConfig(
        type=LogConfig.types.JSON,
        config={
            "max-size": "1g",
        },
    )
    volumes: list | None = None

    port = int(getattr(namespace, "port", "14300"))

    run_command = [
        "uvicorn",
        "pipeline.container.startup:create_app",
        "--host",
        "0.0.0.0",
        "--port",
        str(port),
        "--factory",
    ]

    environment_variables = dict()

    if getattr(namespace, "debug", False):
        run_command.append("--reload")
        current_path = Path("./").expanduser().resolve()
        if volumes is None:
            volumes = []

        volumes.append(f"{current_path}:/app/")
        environment_variables["DEBUG"] = "1"
        environment_variables["LOG_LEVEL"] = "DEBUG"
        environment_variables["FASTAPI_ENV"] = "development"

    if extra_volumes := getattr(namespace, "volume", None):
        if volumes is None:
            volumes = []

        for volume in extra_volumes:
            if ":" not in volume:
                raise ValueError(
                    f"Invalid volume {volume}, must be in format host_path:container_path"  # noqa
                )

            local_path = Path(volume.split(":")[0]).expanduser().resolve()
            container_path = Path(volume.split(":")[1])

            volumes.append(f"{local_path}:{container_path}")

    gpu_ids: list | None = None
    try:
        gpu_ids = [
            f"{i}"
            for i in range(
                0,
                len(
                    subprocess.check_output(
                        [
                            "nvidia-smi",
                            "-L",
                        ]
                    )
                    .decode()
                    .splitlines()
                ),
            )
        ]
    except Exception:
        gpu_ids = None

    # Stop container on python exit
    try:
        container = docker_client.containers.run(
            pipeline_name,
            ports={f"{port}/tcp": int(port)},
            stderr=True,
            stdout=True,
            log_config=lc,
            remove=True,
            auto_remove=True,
            detach=True,
            device_requests=(
                [DeviceRequest(device_ids=gpu_ids, capabilities=[["gpu"]])]
                if gpu_ids
                else None
            ),
            command=run_command,
            volumes=volumes,
            environment=environment_variables,
        )
    except docker.errors.NotFound as e:
        _print(f"Container did not start successfully:\n{e}", "ERROR")
        return

    _print(
        f"Container started on port {port}.\n\n\t\tView the live docs:\n\n\t\t\t http://localhost:{port}/redoc\n\n\t\tor live play:\n\n\t\t\t http://localhost:{port}/play\n",  # noqa
        "SUCCESS",
    )

    while True:
        try:
            for line in container.logs(stream=True):
                print(line.decode("utf-8").strip())
        except KeyboardInterrupt:
            _print("Stopping container...", "WARNING")
            container.stop()
            # container.remove()
            break
        except docker.errors.NotFound:
            _print("Container did not start successfully", "ERROR")
            break



import os
import sys
from datetime import datetime

VERBOSE = bool(int(os.environ.get("VERBOSE", "1")))

LOG_FILE = None


class bcolors:
    PURPLE = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    ORANGE = "\33[38;5;208m"


levels = {
    "WARNING": bcolors.WARNING,
    "INFO": bcolors.PURPLE,
    "ERROR": bcolors.FAIL,
    "REMOTE_LOG": bcolors.OKCYAN,
    "SUCCESS": bcolors.OKGREEN,
}
PIPELINE_STR = f"{bcolors.OKBLUE}Pipeline{bcolors.ENDC}"
PIPELINE_FILE_STR = f"{bcolors.OKBLUE}File{bcolors.ENDC}"


def _print(val, level="INFO"):
    if not VERBOSE:
        return
    time_stamp = datetime.now().strftime("%H:%M:%S")

    log_str = (
        f"{PIPELINE_STR} {time_stamp} - [{levels[level]}{level}{bcolors.ENDC}]: {val}"
    )
    print(f"{log_str}")


def _print_remote_log(val: tuple):
    time_stamp = datetime.fromtimestamp(float(val[0]) / 1e9).strftime("%H:%M:%S.%f")[
        :-3
    ]
    text = val[1]
    normal_string = text

    log_str = f"{PIPELINE_STR} - {bcolors.ORANGE}logs{bcolors.ENDC} {time_stamp}: {normal_string}"  # noqa
    print(f"{log_str}")


def set_print_to_file(path: str):
    global LOG_FILE
    if LOG_FILE is None:
        LOG_FILE = open(path, "w")
        sys.stdout = LOG_FILE
    else:
        raise Exception("Already printing to a log file.")


def stop_print_to_file():
    global LOG_FILE
    if LOG_FILE is not None:
        LOG_FILE.close()
    else:
        raise Exception("Not printing to log file")



import json

from pipeline.cloud import http
from pipeline.cloud.schemas import pointers as pointers_schemas
from pipeline.util.logging import _print


def create_pointer(
    new_pointer: str,
    pointer_or_pipeline_id: str,
    force=False,
) -> None:
    create_schema = pointers_schemas.PointerCreate(
        pointer=new_pointer,
        pointer_or_pipeline_id=pointer_or_pipeline_id,
        locked=False,
    )
    result = http.post(
        "/v4/pointers",
        json.loads(
            create_schema.json(),
        ),
        handle_error=False,
    )

    if result.status_code == 409:
        if force:
            _print("Pointer already exists, forcing update", "WARNING")
            _edit_pointer(new_pointer, pointer_or_pipeline_id)
        else:
            _print(
                f"Pointer {new_pointer} already exists, use --pointer-overwrite to update",  # noqa
                "WARNING",
            )
        return
    elif result.status_code == 201:
        pointer = pointers_schemas.PointerGet.parse_obj(result.json())
        _print(f"Created pointer {pointer.pointer} -> {pointer.pipeline_id}")
    else:
        raise ValueError(f"Failed to create pointer {new_pointer}\n{result.text}")


def _edit_pointer(
    existing_pointer: str,
    pointer_or_pipeline_id: str,
):
    edit_schema = pointers_schemas.PointerPatch(
        pointer_or_pipeline_id=pointer_or_pipeline_id,
        locked=False,
    )

    result = http.patch(
        f"/v4/pointers/{existing_pointer}",
        json.loads(
            edit_schema.json(),
        ),
    )

    if result.status_code == 200:
        pointer = pointers_schemas.PointerGet.parse_obj(result.json())
        _print(f"Updated pointer {pointer.pointer} -> {pointer.pipeline_id}")
    else:
        _print(f"Failed to edit pointer {existing_pointer}", "ERROR")



import json
from argparse import Namespace

import yaml

from pipeline.container import docker_templates
from pipeline.util.logging import _print

from .schemas import PipelineConfig, PythonRuntime, RuntimeConfig


def init_dir(namespace: Namespace) -> None:
    _print("Initializing directory...", "INFO")

    pipeline_name = getattr(namespace, "name", None)

    if not pipeline_name:
        pipeline_name = input("Enter a name for your pipeline: ")

    python_template = docker_templates.pipeline_template_python

    default_config = PipelineConfig(
        runtime=RuntimeConfig(
            container_commands=[
                "apt-get update",
                "apt-get install -y git",
            ],
            python=PythonRuntime(
                version="3.10",
                requirements=[
                    "pipeline-ai",
                ],
            ),
        ),
        accelerators=[],
        pipeline_graph="new_pipeline:my_new_pipeline",
        pipeline_name=pipeline_name,
        accelerator_memory=None,
        extras={},
        readme="README.md",
    )
    with open(getattr(namespace, "file", "./pipeline.yaml"), "w") as f:
        f.write(yaml.dump(json.loads(default_config.json()), sort_keys=False))

    with open("./new_pipeline.py", "w") as f:
        f.write(python_template)

    with open("./README.md", "w") as f:
        f.write(docker_templates.readme_template)

    _print("Initialized directory.", "SUCCESS")



import inspect
import uuid
from hashlib import sha256
from typing import Any, Callable, Dict, Optional

from pipeline.util import generate_id


class Function:
    local_id: str
    remote_id: str

    name: str
    source: str
    hash: str

    typing_inputs: Dict[str, Any]
    typing_outputs: Dict[str, Any]

    function: Callable

    class_instance: Optional[Any]

    def __init__(
        self, function: Callable, *, remote_id: str = None, class_instance: Any = None
    ):
        self.name = function.__name__
        self.remote_id = remote_id
        self.class_instance = class_instance
        self.function = function

        try:
            self.source = inspect.getsource(function)
        except OSError:
            self.source = str(uuid.uuid4())
        self.hash = sha256(self.source.encode()).hexdigest()

        # TODO: Add verification that all inputs to function have a typing annotation,
        # except for "self"
        if "return" not in function.__annotations__:
            function.__annotations__["return"] = None

        self.typing_outputs = {"return": function.__annotations__["return"]}

        self.typing_inputs = {
            function_i: function.__annotations__[function_i]
            for function_i in function.__annotations__
            if not function_i == "return"
        }

        self.local_id = generate_id(10)



import os
import typing as t
from argparse import ArgumentParser, Namespace, _SubParsersAction
from pathlib import Path
from zipfile import ZipFile

import requests
from tabulate import tabulate

from pipeline.cloud import http
from pipeline.cloud.schemas import files as files_schemas
from pipeline.util.logging import _print


def create_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    create_parser = command_parser.add_parser(
        "files",
        aliases=["file"],
        help="Create a new file.",
    )
    create_parser.set_defaults(func=_create_file)

    create_parser.add_argument(
        "path",
        help="Local path to the file.",
    )

    create_parser.add_argument(
        "--name",
        "-n",
        help="Remote alias name of the file.",
        type=str,
    )

    create_parser.add_argument(
        "--recursive",
        "-r",
        help="Recursively upload files in a directory (upload a directory).",
        action="store_true",
    )


def edit_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    ...


def get_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    get_parser = command_parser.add_parser(
        "files",
        aliases=["file"],
        help="Get file information.",
    )
    get_parser.set_defaults(func=_get_file)

    get_parser.add_argument(
        "id",
        nargs="?",
        help="File ID.",
    )

    get_parser.add_argument(
        "--output-file",
        "-o",
        help="Output file path.",
    )

    get_parser.add_argument(
        "--download",
        "-d",
        help="Download file.",
        action="store_true",
    )


def delete_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    ...


def _create_file(args: Namespace) -> None:
    path: str = getattr(args, "path")
    name: str = getattr(args, "name", None)
    recursive: bool = getattr(args, "recursive", False)

    local_path: Path = Path(path)
    local_path = local_path.expanduser().resolve()

    if not local_path.exists():
        raise FileNotFoundError(f"File {local_path} does not exist.")

    if local_path.is_dir() and not recursive:
        raise FileNotFoundError(
            f"File {local_path} is not a file. (Use --recursive to upload a directory.)"
        )

    if recursive:
        tmp_path = Path("/tmp") / (str(local_path.name) + ".zip")
        with ZipFile(str(tmp_path), "w") as zip_file:
            for root, dirs, files in os.walk(str(local_path)):
                for file in files:
                    zip_file.write(
                        os.path.join(root, file),
                        os.path.relpath(os.path.join(root, file), str(local_path)),
                    )
        zip_path = tmp_path

        zip_file = zip_path.open("rb")
        try:
            res = http.post_files(
                "/v3/pipeline_files",
                files=dict(pfile=zip_file),
                progress=True,
            )
        finally:
            zip_file.close()
            print(str(zip_path))
            os.remove(str(zip_path))
        if res is None:
            raise Exception("Failed uploading file")

        if res.status_code != 201:
            raise Exception(
                f"Failed uploading file: {res.text} (code = {res.status_code})"
            )

        res_schema = res.json()
        _print(f"Directory uploaded successfully with ID: {res_schema['id']}")
        return

    query_params = dict()
    if name is not None:
        query_params["name"] = name

    local_file = local_path.open("rb")
    res: requests.Response | None = None
    try:
        res = http.post_files(
            "/v3/pipeline_files",
            files=dict(pfile=local_file),
            progress=True,
            params=query_params,
        )

    finally:
        local_file.close()

    if res is None:
        raise Exception("Failed uploading file")

    if res.status_code != 201:
        raise Exception(f"Failed uploading file: {res.text} (code = {res.status_code})")

    res_schema = res.json()

    _print(f"File uploaded successfully with ID: {res_schema['id']}")


def _get_file(args: Namespace) -> None:
    id: str | None = getattr(args, "id", None)
    output_file: str | None = getattr(args, "output_file", None)
    download: bool = getattr(args, "download", False)

    if download and id is None:
        raise Exception(
            "Cannot download file without ID. (Can only download one file at a time.)"
        )

    if download:
        res = http.get(
            f"/v3/pipeline_files/download/{id}",
        )

        if res.status_code != 200:
            raise Exception(
                f"Failed downloading file: {res.text} (code = {res.status_code})"
            )

        file = res.content
        with open(output_file, "wb") as f:
            f.write(file)

        _print(f"File downloaded successfully to {output_file}")
        return

    if id is not None:
        res = http.get(
            f"/v3/pipeline_files/{id}",
        )

        if res.status_code != 200:
            raise Exception(
                f"Failed getting file: {res.text} (code = {res.status_code})"
            )

        res_schema = res.json()
        res_schema = files_schemas.FileGet.parse_obj(res_schema)

        file = [
            [
                res_schema.id,
                res_schema.created_at,
                res_schema.path,
            ]
        ]

        table = tabulate(
            file,
            headers=[
                "ID",
                "Created At",
                "Path",
            ],
            tablefmt="psql",
        )
        print(table)

        return

    res = http.get(
        "/v3/pipeline_files",
    )

    if res.status_code != 200:
        raise Exception(f"Failed getting files: {res.text} (code = {res.status_code})")

    res_schema: list = res.json()
    res_schema: t.List[files_schemas.FileGet] = [
        files_schemas.FileGet.parse_obj(file_schema) for file_schema in res_schema
    ]

    files = [
        [
            file_schema.id,
            file_schema.created_at,
            file_schema.path,
        ]
        for file_schema in res_schema
    ]

    table = tabulate(
        files,
        headers=[
            "ID",
            "Created At",
            "Path",
        ],
        tablefmt="psql",
    )
    print(table)



import hashlib
import importlib
import os
import traceback
import typing as t
import urllib.parse
from http.client import InvalidURL
from pathlib import Path
from types import NoneType, UnionType
from urllib import request
from urllib.parse import urlparse

from loguru import logger

from pipeline.cloud.schemas import pipelines as pipeline_schemas
from pipeline.cloud.schemas import runs as run_schemas
from pipeline.exceptions import RunInputException, RunnableError
from pipeline.objects import Directory, File, Graph
from pipeline.objects.graph import InputSchema


def is_url(string):
    try:
        result = urlparse(string)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def _get_url_or_path(input_schema: run_schemas.RunInput) -> str | None:
    return input_schema.file_url if input_schema.file_url else input_schema.file_path


class Manager:
    def _load(self, pipeline_path: str):
        with logger.contextualize(pipeline_stage="loading"):
            logger.info("Loading pipeline")

            if ":" not in pipeline_path:
                raise ValueError(
                    "Invalid pipeline path, "
                    + "must be in format <module_or_file>:<pipeline>"
                )
            if len(pipeline_path.split(":")) != 2:
                raise ValueError(
                    "Invalid pipeline path, "
                    + "must be in format <module_or_file>:<pipeline>"
                )

            self.pipeline_path = pipeline_path
            self.pipeline_module_str, self.pipeline_name_str = pipeline_path.split(":")

            self.pipeline_state = pipeline_schemas.PipelineState.loading
            try:
                self.pipeline_module = importlib.import_module(self.pipeline_module_str)

                self.pipeline: Graph = getattr(
                    self.pipeline_module, self.pipeline_name_str
                )
            except ModuleNotFoundError as e:
                raise ValueError(
                    f"Could not load module {self.pipeline_module_str}, {e}"
                )
            except AttributeError:
                raise ValueError(
                    (
                        f"Could not find pipeline {self.pipeline_name_str} in module"
                        f" {self.pipeline_module_str}"
                    )
                )

            self.pipeline_name = os.environ.get("PIPELINE_NAME", "unknown")
            self.pipeline_image = os.environ.get("PIPELINE_IMAGE", "unknown")

            logger.info(f"Pipeline set to {self.pipeline_path}")

    def __init__(self, pipeline_path: str):
        self.pipeline_state: pipeline_schemas.PipelineState = (
            pipeline_schemas.PipelineState.not_loaded
        )
        self.pipeline_state_message: str | None = None
        try:
            self._load(pipeline_path)
        except Exception:
            tb = traceback.format_exc()
            logger.exception("Exception raised when loading pipeline")
            self.pipeline_state = pipeline_schemas.PipelineState.load_failed
            self.pipeline_state_message = tb
            return

    def startup(self):
        if self.pipeline_state == pipeline_schemas.PipelineState.load_failed:
            return
        # add context to enable fetching of startup logs
        with logger.contextualize(pipeline_stage="startup"):
            logger.info("Starting pipeline")
            try:
                self.pipeline._startup()
            except Exception:
                tb = traceback.format_exc()
                logger.exception("Exception raised during pipeline execution")
                self.pipeline_state = pipeline_schemas.PipelineState.startup_failed
                self.pipeline_state_message = tb
            else:
                self.pipeline_state = pipeline_schemas.PipelineState.loaded
                logger.info("Pipeline started successfully")

    def _resolve_file_variable_to_local(
        self,
        file: File | Directory,
        *,
        use_tmp: bool = False,
    ) -> None:
        local_host_dir = "/tmp"

        if hasattr(file, "url") and file.url is not None:
            # Encode the URL to handle spaces and other non-URL-safe characters
            encoded_url = run_schemas.RunInput.encode_url(file.url.geturl())
            cache_name = hashlib.md5(file.url.geturl().encode()).hexdigest()

            file_name = file.url.geturl().split("/")[-1]
            local_path = f"{local_host_dir}/{cache_name}/{file_name}"
            file_path = Path(local_path)
            file_path.parent.mkdir(parents=True, exist_ok=True)
            try:
                # Use the encoded URL for retrieving the file
                request.urlretrieve(encoded_url, local_path)
            # This should not be raise due to encoded_url, but including it in case
            except InvalidURL:
                raise Exception("The file to download has an invalid URL.")
            except Exception:
                raise Exception("Error downloading file.")

        elif file.remote_id is not None or file.path is not None:
            local_path = Path(file.path)
            local_path.parent.mkdir(parents=True, exist_ok=True)
        else:
            raise Exception("File not found, must pass in URL, Path, or Remote ID.")

        if isinstance(file, Directory):
            raise NotImplementedError("Remote ID not implemented yet")

        file.path = Path(local_path)

    def _create_file_variable(
        self,
        *,
        path_or_url: str | None = None,
        is_directory: bool = False,
        use_tmp: bool = False,
    ) -> File | Directory:
        path: str | None = None
        url: str | None = None

        if is_url(path_or_url):
            url = str(urllib.parse.urlparse(path_or_url).geturl())
        else:
            path = path_or_url

        if path is None and url is None:
            raise Exception("Must provide either path or url")

        variable: File | Directory | None = None

        if is_directory:
            variable = Directory(path=path, url=url)
        else:
            variable = File(path=path, url=url)

        self._resolve_file_variable_to_local(variable, use_tmp=use_tmp)
        return variable

    def _parse_inputs(
        self, input_data: t.List[run_schemas.RunInput] | None, graph: Graph
    ) -> t.List[t.Any]:
        inputs = []
        if input_data is None:
            input_data = []
            return input_data
        if len(input_data) > 0:
            for item in input_data:
                input_schema = run_schemas.RunInput.parse_obj(item)
                if input_schema.type == run_schemas.RunIOType.file:
                    if input_schema.file_path is None and input_schema.file_url is None:
                        raise RunInputException(
                            "A file must either have a path or url attribute"
                        )
                    path_or_url = _get_url_or_path(input_schema)
                    if path_or_url is None:
                        raise RunInputException(
                            "A file must either have a path or url attribute"
                        )

                    variable = self._create_file_variable(path_or_url=path_or_url)
                    inputs.append(
                        variable,
                    )

                else:
                    inputs.append(input_schema.value)
        graph_inputs = list(filter(lambda v: v.is_input, graph.variables))

        if len(inputs) != len(graph_inputs):
            raise RunInputException("Inputs do not match graph inputs")

        final_inputs = []

        for i, (user_input, model_input) in enumerate(zip(inputs, graph_inputs)):
            target_type = model_input.type_class
            if issubclass(target_type, InputSchema) and isinstance(user_input, dict):
                # schema_instance = target_type()
                annotations = target_type.__annotations__
                for key, value in annotations.items():
                    if isinstance(value, UnionType) or "typing.Optional" in str(value):
                        var_union_types = list(t.get_args(value))
                        if len(var_union_types) > 2:
                            raise RunInputException("Only support Union of 2 types")
                        if NoneType in var_union_types:
                            var_union_types.remove(NoneType)
                        else:
                            raise RunInputException("Only support Union with None")
                        var_type = var_union_types[0]
                    else:
                        var_type = value

                    if var_type == File:
                        if user_input.get(key) is None:
                            continue
                        file_schema = run_schemas.RunInput.parse_obj(user_input[key])
                        path_or_url = _get_url_or_path(file_schema)
                        variable = self._create_file_variable(
                            path_or_url=path_or_url,
                            use_tmp=True,
                        )
                        user_input[key] = variable

            final_inputs.append(user_input)
        return final_inputs

    def run(
        self, run_id: str | None, input_data: t.List[run_schemas.RunInput] | None
    ) -> t.Any:
        with logger.contextualize(run_id=run_id):
            logger.info("Running pipeline")
            args = self._parse_inputs(input_data, self.pipeline)
            try:
                result = self.pipeline.run(*args)
            except RunInputException:
                raise
            except Exception as exc:
                raise RunnableError(exception=exc, traceback=traceback.format_exc())
            logger.info("Run successful")
            return result



import inspect
from pathlib import Path
from types import NoneType, UnionType
from typing import Any, Generic, Iterable, List, TypeVar, get_args
from urllib.parse import ParseResult, urlparse

import httpx
from cloudpickle import dumps, loads
from tqdm import tqdm

from pipeline.cloud.schemas.pipelines import IOVariable
from pipeline.cloud.schemas.runs import RunIOType
from pipeline.exceptions import RunInputException
from pipeline.objects.function import Function
from pipeline.objects.model import Model
from pipeline.util import generate_id


class InputSchema:
    def __init__(self, **kwargs):
        for key, value in self.__annotations__.items():
            validation_field = getattr(self, key, None)
            if not isinstance(validation_field, InputField):
                raise Exception("Must be InputField")

            if key not in kwargs and (
                "typing.Optional" in str(value) or isinstance(value, UnionType)
            ):
                setattr(self, key, validation_field.default)
                continue

            if key not in kwargs and not (
                "typing.Optional" in str(value) or isinstance(value, UnionType)
            ):
                raise Exception(
                    f"Missing value for '{key}', if you want to make it optional, use 'typing.Optional' or the pipe operator for example: 'int | None'"  # noqa
                )

            validation_field.validate(kwargs[key])
            setattr(self, key, kwargs[key])

    def __repr__(self) -> str:
        vars = ", ".join(
            [f"{key}={getattr(self, key)}" for key in self.__annotations__.keys()]
        )
        return f"InputSchema({vars})"

    @classmethod
    def to_schema(cls) -> List[dict]:
        output_list: List[IOVariable] = []

        items = dir(cls)

        for item, title in [
            (_field, item)
            for item in items
            if (_field := getattr(cls, item, None)) and isinstance(_field, InputField)
        ]:
            var_type = cls.__annotations__[title]

            if isinstance(var_type, UnionType) or "typing.Optional" in str(var_type):
                var_union_types = list(get_args(var_type))
                if len(var_union_types) > 2:
                    raise Exception("Only support Union of 2 types")
                if NoneType in var_union_types:
                    var_union_types.remove(NoneType)
                else:
                    raise Exception("Only support Union with None")
                var_type = var_union_types[0]

            output_list.append(
                item._to_io_schema(
                    _type=var_type,
                    _title=title,
                ).dict()
            )
        return output_list

    @classmethod
    def from_schema(cls, schemas: List[dict]) -> "InputSchema":
        new_input_schema = cls()
        new_input_schema.__annotations__ = {}
        for schema in schemas:
            input_field = InputField._from_io_schema(IOVariable(**schema))

            setattr(
                new_input_schema,
                schema.get("title"),
                input_field,
            )

            new_input_schema.__annotations__[schema.get("title")] = RunIOType.to_object(
                schema.get("run_io_type")
            )

        return new_input_schema

    def parse_dict(self, input_dict: dict) -> None:
        for item, title in [
            (_field, item)
            for item in dir(self)
            if (_field := getattr(self, item, None)) and isinstance(_field, InputField)
        ]:
            print(f"Assesing {title}")
            entered_value = input_dict.get(title, None)
            if item.default is None and entered_value is None:
                raise ValueError(f"Field {title} is not optional, no value entered")

            item.validate(entered_value)

            print(f"Now {title}={entered_value}")
            setattr(self, title, entered_value)

    def to_dict(self) -> dict:
        return {key: getattr(self, key) for key in self.__annotations__.keys()}


class A:
    pass


T: Any = A


class InputField(T):
    def __init__(
        self,
        default: Any | None = None,
        title: str | None = None,
        description: str | None = None,
        examples: list[Any] | None = None,
        gt: int | None = None,
        ge: int | None = None,
        lt: int | None = None,
        le: int | None = None,
        multiple_of: int | None = None,
        allow_inf_nan: bool | None = None,
        max_digits: int | None = None,
        decimal_places: int | None = None,
        min_length: int | None = None,
        max_length: int | None = None,
        choices: list[Any] | None = None,
        optional: bool | None = False,
    ):
        self.default = default
        self.title = title
        self.description = description
        self.examples = examples
        self.gt = gt
        self.ge = ge
        self.lt = lt
        self.le = le
        self.multiple_of = multiple_of
        self.allow_inf_nan = allow_inf_nan
        self.max_digits = max_digits
        self.decimal_places = decimal_places
        self.min_length = min_length
        self.max_length = max_length
        self.choices = choices
        self.optional = optional

        if default is not None:
            try:
                self.validate(self.default)
            except TypeError as e:
                raise TypeError(
                    f"Default value {default} is invalid for field {self.title}"
                ) from e

    def _to_io_schema(self, _type: Any, _title: str) -> IOVariable:
        return IOVariable(
            run_io_type=RunIOType.from_object(_type),
            title=_title,
            description=self.description,
            examples=self.examples,
            gt=self.gt,
            ge=self.ge,
            lt=self.lt,
            le=self.le,
            multiple_of=self.multiple_of,
            allow_inf_nan=self.allow_inf_nan,
            max_digits=self.max_digits,
            decimal_places=self.decimal_places,
            min_length=self.min_length,
            max_length=self.max_length,
            choices=self.choices,
            default=self.default,
            optional=self.optional,
        )

    @classmethod
    def _from_io_schema(cls, schema: IOVariable):
        return cls(
            default=schema.default,
            description=schema.description,
            examples=schema.examples,
            gt=schema.gt,
            ge=schema.ge,
            lt=schema.lt,
            le=schema.le,
            multiple_of=schema.multiple_of,
            allow_inf_nan=schema.allow_inf_nan,
            max_digits=schema.max_digits,
            decimal_places=schema.decimal_places,
            min_length=schema.min_length,
            max_length=schema.max_length,
            choices=schema.choices,
            title=schema.title,
            optional=schema.optional,
        )

    def validate(self, value: Any):
        if self.choices is not None:
            if value not in self.choices:
                raise TypeError(f"Value is not in the choices of {self.choices}")

        if self.gt is not None:
            if value <= self.gt:
                raise TypeError(f"Value is not greater than {self.gt}")

        if self.ge is not None:
            if value < self.ge:
                raise TypeError(f"Value is not greater than or equal to {self.ge}")

        if self.lt is not None:
            if value >= self.lt:
                raise TypeError(f"Value is not less than {self.lt}")

        if self.le is not None:
            if value > self.le:
                raise TypeError(f"Value is not less than or equal to {self.le}")

        if self.multiple_of is not None:
            if value % self.multiple_of != 0:
                raise TypeError(f"Value is not a multiple of {self.multiple_of}")

        if self.allow_inf_nan is not None:
            if not self.allow_inf_nan and (
                value == float("inf") or value == float("-inf") or value == float("nan")
            ):
                raise TypeError("Value is not allowed to be infinity or nan")

        if self.max_digits is not None:
            if len(str(value)) > self.max_digits:
                raise TypeError(f"Value has more than {self.max_digits} digits")

        if self.decimal_places is not None:
            if len(str(value).split(".")[1]) > self.decimal_places:
                raise TypeError(
                    f"Value has more than {self.decimal_places} decimal places"
                )

        if self.min_length is not None:
            if len(str(value)) < self.min_length:
                raise TypeError(f"Value has less than {self.min_length} characters")

        if self.max_length is not None:
            if len(str(value)) > self.max_length:
                raise TypeError(f"Value has more than {self.max_length} characters")


class Variable:
    local_id: str | None
    remote_id: str | None

    name: str | None

    type_class: Any

    is_input: bool
    is_output: bool

    def __init__(
        self,
        type_class: Any,
        *,
        default: Any | None = None,
        is_input: bool = True,
        is_output: bool = False,
        local_id: str | None = None,
        # default: Any | None = None, # TODO: Implement default values
        title: str | None = None,
        description: str | None = None,
        examples: list[Any] | None = None,
        gt: int | None = None,
        ge: int | None = None,
        lt: int | None = None,
        le: int | None = None,
        multiple_of: int | None = None,
        allow_inf_nan: bool | None = None,
        max_digits: int | None = None,
        decimal_places: int | None = None,
        min_length: int | None = None,
        max_length: int | None = None,
        choices: list[Any] | None = None,
        allow_out_of_context_creation: bool = False,
    ):
        from pipeline.objects.pipeline import Pipeline

        self.type_class = type_class
        self.is_input = is_input
        self.is_output = is_output

        self.title = title
        self.description = description
        self.examples = examples
        self.gt = gt
        self.ge = ge
        self.lt = lt
        self.le = le
        self.multiple_of = multiple_of
        self.allow_inf_nan = allow_inf_nan
        self.max_digits = max_digits
        self.decimal_places = decimal_places
        self.min_length = min_length
        self.max_length = max_length
        self.choices = choices
        self.default = default
        self.dict_schema = (
            type_class.to_schema()
            if inspect.isclass(type_class)
            and (
                issubclass(type_class, InputSchema)
                or isinstance(type_class, InputSchema)
            )
            else None
        )
        if not Pipeline._pipeline_context_active and not allow_out_of_context_creation:
            raise Exception("Cant add a variable when not defining a pipeline")

        if Pipeline._pipeline_context_active:
            Pipeline._current_pipeline.variables.append(self)

        self.local_id = generate_id(10) if not local_id else local_id

    def validate_variable(self, value: Any) -> None:
        # if not isinstance(value, self.type_class):
        #     return

        if self.choices is not None:
            if value not in self.choices:
                raise TypeError(f"Value is not in the choices of {self.choices}")

        if self.gt is not None:
            if value <= self.gt:
                raise TypeError(f"Value is not greater than {self.gt}")

        if self.ge is not None:
            if value < self.ge:
                raise TypeError(f"Value is not greater than or equal to {self.ge}")

        if self.lt is not None:
            if value >= self.lt:
                raise TypeError(f"Value is not less than {self.lt}")

        if self.le is not None:
            if value > self.le:
                raise TypeError(f"Value is not less than or equal to {self.le}")

        if self.multiple_of is not None:
            if value % self.multiple_of != 0:
                raise TypeError(f"Value is not a multiple of {self.multiple_of}")

        if self.allow_inf_nan is not None:
            if not self.allow_inf_nan and (
                value == float("inf") or value == float("-inf") or value == float("nan")
            ):
                raise TypeError("Value is not allowed to be infinity or nan")

        if self.max_digits is not None:
            if len(str(value)) > self.max_digits:
                raise TypeError(f"Value has more than {self.max_digits} digits")

        if self.decimal_places is not None:
            if len(str(value).split(".")[1]) > self.decimal_places:
                raise TypeError(
                    f"Value has more than {self.decimal_places} decimal places"
                )

        if self.min_length is not None:
            if len(str(value)) < self.min_length:
                raise TypeError(f"Value has less than {self.min_length} characters")

        if self.max_length is not None:
            if len(str(value)) > self.max_length:
                raise TypeError(f"Value has more than {self.max_length} characters")

    def to_io_schema(self) -> IOVariable:
        return IOVariable(
            run_io_type=(
                RunIOType.dictionary
                if (
                    inspect.isclass(self.type_class)
                    and issubclass(self.type_class, InputSchema)
                )
                else RunIOType.from_object(self.type_class)
            ),
            title=self.title,
            description=self.description,
            examples=self.examples,
            gt=self.gt,
            ge=self.ge,
            lt=self.lt,
            le=self.le,
            multiple_of=self.multiple_of,
            allow_inf_nan=self.allow_inf_nan,
            max_digits=self.max_digits,
            decimal_places=self.decimal_places,
            min_length=self.min_length,
            max_length=self.max_length,
            choices=self.choices,
            dict_schema=self.dict_schema,
            # Backwards compatible with older pipelines
            default=self.default if hasattr(self, "default") else None,
        )


class File(Variable):
    path: Path | None
    url: ParseResult | None
    remote_id: str | None

    def __init__(
        self,
        *,
        path: str | Path | None = None,
        remote_id: str | None = None,
        url: str | None = None,
        title: str | None = None,
        allow_out_of_context_creation: bool = True,
        local_id: str | None = None,
    ) -> None:
        super().__init__(
            type_class=self.__class__,
            is_input=False,
            is_output=False,
            title=title,
            local_id=local_id,
            allow_out_of_context_creation=allow_out_of_context_creation,
        )

        self.path = (
            path
            if isinstance(path, Path)
            else (Path(path) if path is not None else None)
        )
        self.remote_id: str | None = remote_id
        self.url = urlparse(url) if url is not None else None

    def save(self, path: str | Path) -> None:
        if self.path is None and self.url is None:
            raise Exception("Path and URL are None")

        if isinstance(path, str):
            path = Path(path)

        if self.path is not None and self.path.is_dir():
            raise Exception("Path is a directory")

        if self.url is not None:
            with path.open("wb") as f:
                with httpx.stream("GET", self.url.geturl()) as response:
                    total = int(response.headers["Content-Length"])

                    with tqdm(
                        total=total, unit_scale=True, unit_divisor=1024, unit="B"
                    ) as progress:
                        num_bytes_downloaded = response.num_bytes_downloaded
                        for chunk in response.iter_bytes():
                            f.write(chunk)
                            progress.update(
                                response.num_bytes_downloaded - num_bytes_downloaded
                            )
                            num_bytes_downloaded = response.num_bytes_downloaded

        elif self.path is not None:
            path.write_bytes(self.path.read_bytes())


class Directory(File):
    def __init__(
        self,
        *,
        path: str | Path | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            path=path,
            **kwargs,
        )

        if (
            self.path is not None
            and not self.path.is_dir()
            and not str(self.path).endswith(".zip")
            and self.remote_id is None
        ):
            raise Exception("Path is not a directory")

    @classmethod
    def from_object(
        cls,
        *args,
        **kwargs,
    ):
        raise NotImplementedError("Directory.from_object is not implemented")


ST = TypeVar("ST")


class Stream(Generic[ST], Iterable):
    def __init__(self, iterable: Iterable[ST]):
        self.iterable = iterable

    def __iter__(self):
        return self.iterable.__iter__()

    def __next__(self):
        return self.iterable.__next__()


class GraphNode:
    local_id: str
    function: Function
    inputs: List[Variable] = []
    outputs: List[Variable] = []

    def __init__(self, function, inputs, outputs, *, local_id=None):
        self.function = function
        self.inputs = inputs
        self.outputs = outputs

        self.local_id = generate_id(10) if local_id is None else local_id


class Graph:
    local_id: str
    remote_id: str

    name: str

    functions: List[Function]
    variables: List[Variable]

    outputs: List[Variable]

    nodes: List[GraphNode]

    models: List[Model]

    def __init__(
        self,
        *,
        name: str = "",
        variables: List[Variable] = None,
        functions: List[Function] = None,
        outputs: List[Variable] = None,
        nodes: List[GraphNode] = None,
        models: List[Model] = None,
    ):
        self.name = name
        self.local_id = generate_id(10)

        self.variables = variables if variables is not None else []
        self.functions = functions if functions is not None else []
        self.outputs = outputs if outputs is not None else []
        self.nodes = nodes if nodes is not None else []
        self.models = models if models is not None else []
        # Flag set when all functions with the on_startup field have run
        self._has_run_startup = False

    def _startup(self):
        if self._has_run_startup:
            return

        startup_variables = {}

        for var in self.variables:
            # At the moment only the File variable can be used on startup
            if isinstance(var, File):
                startup_variables[var.local_id] = var

        for node in self.nodes:
            node_inputs: List[Variable] = []
            node_function: Function = None

            for function in self.functions:
                if function.local_id == node.function.local_id:
                    node_function = function
                    break
            if (
                getattr(node_function.function, "__run_once__", False)
                and getattr(node_function.function, "__has_run__", False)
            ) or not getattr(node_function.function, "__on_startup__", False):
                continue

            for _node_input in node.inputs:
                for variable in self.variables:
                    if variable.local_id == _node_input.local_id:
                        node_inputs.append(variable)
                        break

            function_inputs = []
            for _input in node_inputs:
                function_inputs.append(startup_variables[_input.local_id])

            if node_function.function is None:
                raise Exception(
                    "Node function is None (id:%s)" % node.function.local_id
                )

            if getattr(node_function, "class_instance", None) is not None:
                node_function.function(node_function.class_instance, *function_inputs)
            else:
                node_function.function(*function_inputs)

            if getattr(node_function.function, "__run_once__", False):
                node_function.function.__has_run__ = True

        self._has_run_startup = True

    def run(self, *inputs):
        input_variables: List[Variable] = [
            var for var in self.variables if var.is_input
        ]

        if len(inputs) != len(input_variables):
            raise RunInputException(
                "Mismatch of number of inputs, expecting %u got %s"
                % (len(input_variables), len(inputs))
            )

        self._startup()

        running_variables = {}

        # Add all File's to the running variables
        for var in self.variables:
            if isinstance(var, File):
                if not var.path:
                    raise RunInputException("Must define a path for a File")

                running_variables[var.local_id] = var

        for i, input in enumerate(inputs):
            target_type = input_variables[i].type_class

            if issubclass(target_type, InputSchema) and isinstance(input, dict):
                input = target_type(**input)
            elif not isinstance(input, input_variables[i].type_class):
                if isinstance(input, int) and input_variables[i].type_class == float:
                    input = float(input)
                else:
                    raise RunInputException(
                        "Input type mismatch, expected %s got %s"
                        % (
                            input_variables[i].type_class,
                            input.__class__,
                        )
                    )
            running_variables[input_variables[i].local_id] = input

        for node in self.nodes:
            node_inputs: List[Variable] = []
            node_outputs: List[Variable] = []
            node_function: Function = None

            for function in self.functions:
                if function.local_id == node.function.local_id:
                    node_function = function
                    break
            if getattr(node_function.function, "__run_once__", False) and getattr(
                node_function.function, "__has_run__", False
            ):
                continue

            for _node_input in node.inputs:
                for variable in self.variables:
                    if variable.local_id == _node_input.local_id:
                        node_inputs.append(variable)
                        break
            for _node_output in node.outputs:
                for variable in self.variables:
                    if variable.local_id == _node_output.local_id:
                        node_outputs.append(variable)
                        break

            function_inputs = []
            for _input in node_inputs:
                function_inputs.append(running_variables[_input.local_id])

            if node_function.function is None:
                raise Exception(
                    "Node function is none (id:%s)" % node.function.local_id
                )

            if getattr(node_function, "class_instance", None) is not None:
                output = node_function.function(
                    node_function.class_instance, *function_inputs
                )
            else:
                output = node_function.function(*function_inputs)

            if len(node.outputs) > 1:
                if len(node.outputs) == len(output) == len(node_outputs):
                    number_of_outputs = len(output)
                    for i in range(number_of_outputs):
                        running_variables[node_outputs[i].local_id] = output[i]
                else:
                    raise Exception(
                        "Mismatch in number of outputs:"
                        f"{len(node.outputs)}/{len(output)}/{len(node_outputs)}"
                    )
            else:
                running_variables[node_outputs[0].local_id] = output

            if not getattr(node_function.function, "__has_run__", False):
                node_function.function.__has_run__ = True

        return_variables = []

        for output_variable in self.outputs:
            return_variables.append(running_variables[output_variable.local_id])

        return return_variables

    def save(self, save_path):
        with open(save_path, "wb") as save_file:
            save_file.write(dumps(self))

    @classmethod
    def load(cls, load_path):
        with open(load_path, "rb") as load_file:
            return loads(load_file.read())
